//note: next line is for validation system
/*global gController, MinimalController, AlgoLabModel, DynamicArray,
newline, TextStream, AlgoMethodSet
*/

/*TODO
need to validate for dot operator, e.g.,
	foo.	okay
	10.		no
*/

//algo-lab-controller.js
//class: AlgoLabController
//algo = algorithm

/*
NOTE WELL --
This script is intended for dev.
I.e., for running locally (sandboxing).
We would not want to put into production because we use Function with UI input.
*/

//Initializing

///* eslint no-undef: ["error", { "varsIgnorePattern": "gController" }] */

// eslint-disable-next-line no-unused-vars
window.onerror = function (msg, url, lineNo, columnNo, error) {
	if (msg.includes('Script error'))
		alert(`Error loading a script. After browser opens, press "F12" or Ctrl+Shift+I and look on "console" for load errors (script filename and line number)`);
};


let gTypeAssistant = null;

/**
 * initialization when window loading is "complete"
 */
document.onreadystatechange = function () {
	if (document.readyState === 'complete')
		try {
			buildController();
			gController.opened();
			gController.initDotAssistant();
			//Special caching behavior to allow preservation of source
			//files while reloading source files
			const labInputText = sessionStorage.getItem('labInputText');
			if (labInputText) {
				sessionStorage.removeItem('labInputText');
				if (labInputText.trim().length > 0) {
					gController.setInputText(labInputText);
					gController.changedInput();
				}
			}
			//for some crazy reason, it's defaulting scrolled to right (we fix that here)
			gController.inputElem().scrollLeft = 0;
		} catch (ex) {
			const consoleNote = 'See console for more information';
			alert('*Exception* \n' + ex.message + '\n' + consoleNote);
		}
}

//-------------------------------------------------------------

class AlgoLabController extends MinimalController {

	constructor() {
		super();
		//TODO -- hard coded classes here
		//this.model(new AlgoLabModel([MyArray, MyString, MyDate]));
		this.model(new AlgoLabModel([DynamicArray]));
	}

	model(arg) { if (arg !== undefined) this._model = arg; return this._model; }

	opened() {

		//action hookup
		this.attach('reload', 'click', 'clickedReload');
		this.attach('run', 'click', 'clickedRun');
		this.attach('clear', 'click', 'clickedClear');
		this.attach('copyUp', 'click', 'clickedCopyUp');
		this.attach('copyAll', 'click', 'clickedCopyAll');

		//change events
		this.attach('static', 'change', 'changedStatic');
		this.attach('instance', 'change', 'changedInstance');

		//source (input) events
		this.attach('input', 'change', 'changedInput');
		this.attach('input', 'select', 'selectionChangedInput');
		this.attach('input', 'blur', 'losingFocusInput');
		this.attach('input', 'focus', 'gettingFocusInput');
		this.attach('input', 'focus', 'gettingFocusInput');
		this.attach('input', 'keydown', 'keydownInput');
		this.attach('input', 'paste', 'pasteInput');
		this.basicAttachEventToHandler(this.inputElem(), "keyup", 'keyupInput');

		//This is important for initial (cursor) behavior in the textarea field
		//i.e., replacing undefined value with empty string
		//if (!this.inputElem().value))
		//	this.inputElem().value = "";
		this.attachKeyupEvent();
		this.attachKeydownEvent();

		this.attach('classes', 'input', 'changedClasses');
		this.attach('methods', 'input', 'changedMethods');

		this.buildClassList();
	}

	//-------------------------------------------------------------------
	//Event Handlers

	gettingFocusInput() {
		this.setSelectedInputText(this.model().selectionRange());
		//this.setSelectedInputText(this.model().selectedText());
	}

	changedInput() {
		let src = this.inputElem().value;
		let lines = src.split(/\r?\n/);
		lines = lines.map(line => line.trimRight());
		src = lines.join(newline());
		this.inputElem().value = src;
		this.model().source(src);
	}

	applyInput() {
		this.model().source(this.inputElem().value);
	}

	selectionChangedInput() {
		this.model().selectionRange(this.getSelectedInputTextPair());
	}

	changedInstance() {
		const value = this.getElem('instance').value;
		if (value === 'instance') {
			this.model().instanceOption(true);
			this.buildMethodsList();
		}
	}

	changedStatic() {
		const value = this.getElem('static').value;
		if (value === 'static') {
			this.model().instanceOption(false);
			this.buildMethodsList();
		}
	}

	losingFocusInput() {
		//this.model().source(this.inputElem().value);
		//this.model().selectionRange(this.getSelectedInputTextPair());
	}

	pasteInput(event) {
		let pastedString, unindentedString;
		if (this.inputElem().value.trim())
			return;
		event.preventDefault();
		pastedString = (event.clipboardData || window.clipboardData).getData('text');
		unindentedString = TextStream.unindent(pastedString);
		event.preventDefault()
		this.setInputText(unindentedString);
		this.updateModel();
	}

	//--------------------------------------
	//Getting Elements

	inputElem() {
		return this.getElem('input');
	}

	outputElem() {
		return this.getElem('output');
	}

	//--------------------------------------
	//Actions

	clickedRun() {
		//try
		//this.triggerFocusedInputChange();
		this.changedInput();

		this.clearErrorText();
		this.model().selectionRange(this.getSelectedInputTextPair());
		const error = this.model().run();
		if (error && error.hasError())
			this.applyInputError(error);
	}

	clickedClear() {
		this.outputElem().value = "";
	}

	clickedCopyUp() {
		this.triggerFocusedInputChange();
		let elem, value, newText;
		elem = this.inputElem();
		value = elem.value;
		newText = this.fCodeSource().value;
		newText = this.constructor.parseExample(newText);
		value += '\n' + newText;
		elem.value = value;
	}

	static parseExample(text) {
		//Assume example is after line beginning with "e.g."
		if (!text.includes('e.g.')) return;
		const strm = TextStream.from(text);
		if (strm.moveTo(/^\s*e\.g\./)) {
			strm.sw();
			strm.nextLine();
			strm.sw();
			return strm.copyRemaining();
		}
		return text;
	}

	clickedCopyAll() {
		// eslint-disable-next-line no-undef
		copyToClipboard(this.inputElem().value.trim());
	}

	clickedReload() {
		this.triggerFocusedInputChange();
		this.clearErrorText();
		sessionStorage.setItem("labInputText", this.inputElem().value);
		document.location.reload();
	}

	/*appendInputError(error) {
		let text;
		text = this.inputElem().value.trimRight();
		text += '\n\n//' + error;
		this.inputElem().value = text;
	}*/

	//-------------------------------------------------------------------
	//Handling and routing enter key to "run" action

	keyupWindow(evt) {
		// "Enter" key is 13 on keyboard
		/* problem is enter on "Assistant" kicks in here
		if (evt.target.id !== "input" && evt.keyCode === 13) {
			evt.preventDefault();
			this.clickedRun();
		}*/
		// "F9" key is 120
		if (evt.keyCode === 120) {
			evt.preventDefault();
			this.clickedRun();
		}
		// "F4" key is 115
		if (evt.keyCode === 115) {
			evt.preventDefault();
			this.clickedClear();
		}
	}

	keydownWindow(evt) {
		// "F5" key is 116
		if (evt.keyCode === 116) {
			evt.preventDefault();
			//alert('F5 wowee')
			//this.clickedRun();
			this.clickedReload();
		}
	}

	attachKeyupEvent() {
		this.basicAttachEventToHandler(window, "keyup", 'keyupWindow');
	}

	attachKeydownEvent() {
		this.basicAttachEventToHandler(window, "keydown", 'keydownWindow');
	}

	hitRefresh() {
		sessionStorage.setItem("labInputText", this.inputElem().value);
	}

	//-------------------------------------------------------------------
	//Input Field Dot Operator Handling

	keyupInput(evt) {
		// "." key is 190 (some list as 46??)
		//helpful mapping: https://developer.mozilla.org/en-US/docs/Web/API/KeyboardEvent/keyCode
		if (evt.keyCode === 190) {
			evt.preventDefault();
			this.clickedDotOperator(evt);
		}
	}

	keydownInput(evt) {
		//tab=9
		//helpful mapping: https://keycode.info/
		//NOTE XDOC -- seems odd keyup did not work for tab (handler is not called), keydown did
		if (evt.keyCode === 9) {
			evt.preventDefault();
			const field = this.inputElem();
			const p = field.selectionStart;
			field.value = field.value.slice(0, p) +
				'	' + field.value.slice(field.selectionStart);
			const p2 = field.selectionStart = p + 1;
			field.selectionStart = p2;
			field.selectionEnd = p2;
		}
	}

	updateModel() {
		this.applyInput();
		this.selectionChangedInput();
	}

	clickedDotOperator(evt) {
		//update model
		let m, sanity, labSource, type, cl, methodSet, methodList;
		this.updateModel();
		m = this.model();
		//todo -- remoev drillig ..
		labSource = m.labSource();
		const cp = labSource.cursorPos();
		labSource.setCursorPos(cp - 1);
		sanity = labSource.currentCharacter();
		if (sanity !== '.')
			throw 'expecting dot';
		type = labSource.typeAtDotOperator();
		if (!type) { this.selectionChangedInput(); return; }
		cl = m.zclassChoicesMap().get(type);
		if (!cl) { this.selectionChangedInput(); return; }
		methodSet = new AlgoMethodSet(cl);
		methodList = methodSet.methodHeaders();
		methodList = methodList.map(ea => ea.trim());

		gTypeAssistant.setMenu(methodList);
		gTypeAssistant.open(evt);

		/*
		alert(methodList.join('\n'));
		m.insert('indexOf(anIndex)');
		this.refreshInput();
		const cpNew = cp + 'indexOf(anIndex)'.length;
		this.setSelectedInputText([cpNew, cpNew]);
		*/
	}

	//-------------------------------------------------------------------
	// Refreshing Input Ui From Model

	refreshInput() {
		this.setInputText(this.model().getSource());
		this.setSelectedInputText(this.model().selectionRange());
	}

	//-------------------------------------------------------------------
	//Class list

	buildClassList() {
		const field = this.fClasses();
		const classes = this.model().zclassChoices();
		classes.forEach(aClass => {
			const classNm = aClass.name;
			let item = document.createElement("option");
			item.value = classNm;
			item.innerText = classNm;
			field.appendChild(item);
		});
		if (field.options.length > 0) {
			field.options[0].selected = true;
			this.model().zclass(field.value);
			this.changedClasses();
		}
	}

	changedClasses() {
		this.buildMethodsList(this.fClasses().value);
	}

	//-------------------------------------------------------------------
	//Methods list

	static methodHeaderFrom(method) {
		//Answer method name and arg names
		//foo(w, h)
		const right = ")";
		return method.toString().split(right)[0] + right;
	}

	buildMethodsList(selectedClass) {
		const listField = this.fMethods();
		this.clearMethodsList();
		this.model().zclass(selectedClass);
		this.model().methodHeaders().forEach(methodHeader => {
			let item;
			item = document.createElement("option");
			item.value = methodHeader;
			item.innerText = methodHeader;
			listField.appendChild(item);
		});
		if (listField.options.length > 0) {
			listField.options[0].selected = true;
			this.model().zmethod(listField.value);
		}
		this.changedMethods();
	}

	clearMethodsList() {
		this.fMethods().innerHTML = '';	//clears methods
	}

	changedMethods() {
		this.updateMethodComment();
	}

	//-------------------------------------------------------------------
	//Method comment (for selected method)

	updateMethodComment() {
		const
			m = this.model(),
			currentMethod = this.coerceMethodName(this.fMethods().value);
		m.zmethod(currentMethod);
		const comment = m.selectedMethodComment();
		this.fCodeSource().value = comment;
		//this.fCodeSource().innerText = comment;
	}

	//-------------------------------------------------------------------
	//Helpers

	coerceMethodName(arg) {
		//arg may be method name or method header
		let i = arg.indexOf('(');
		if (i < 0) i = arg.length;
		return arg.slice(0, i);
	}

	getSelectedInputText() {
		//return selected text (may be empty string)
		const
			elem = this.inputElem(),
			text = elem.value,
			pair = this.getSelectedInputTextPair(),
			i1 = (pair ? pair[0] : 0),
			i2 = (pair ? pair[1] : 0);
		if (i1 != null && i2 != null)
			return text.substring(i1, i2);
		else
			return "";
	}

	getSelectedInputTextPair() {
		//return selected text (may be empty string)
		return [this.inputElem().selectionStart, this.inputElem().selectionEnd];
	}

	setSelectedInputText(pair) {
		//TODO -- this method has confusing name
		//return selected text (may be empty string)
		if (!pair) return;
		if (!Array.isArray(pair))
			throw 'setSelectedInputText expects array';
		this.inputElem().focus();
		this.inputElem().setSelectionRange(parseInt(pair[0]), parseInt(pair[1]));
	}

	setInputText(text) {
		this.inputElem().value = text;
		this.setSelectedInputText([0, 0]);
	}

	setErrorText(text) {
		this.elem('error').innerText = text;
	}

	clearErrorText() {
		this.setErrorText('');
	}

	triggerFocusedInputChange() {
		//if input has focus, trigger change
		//this is for cases e.g. when user is in field and hits accelerator key
		//browser does not send change event in this case
		if (document.activeElement === this.inputElem())
			this.changedInput();
	}

	applyInputError(error) {
		this.setErrorText(error.getMessage());
		this.setSelectedInputText(this.model().errorRangeFor(error));
	}

	//-------------------------------------------------------------------
	// Dot Operator Assistant in Input Area
	// (popup menu with candidate types)

	initDotAssistant() {
		// eslint-disable-next-line no-undef
		gTypeAssistant = PopupMenu.fromTargetMenuResult(this, this.inputElem(), [], this.dotAssistantResultFct);
		gTypeAssistant.registerEvents();
	}

	dotAssistantResultFct(aType) {
		this.model().insert(aType);
		this.refreshInput();
		const cp = this.model().labSource().cursorPos();
		const cpNew = cp + aType.length + 1;
		this.setSelectedInputText([cpNew, cpNew]);
	}

	//-------------------------------------------------------------------
	//Fields/Panes
	//fClasses means the field element 'classes'

	fClasses() { return this.getElem('classes'); }
	fMethods() { return this.getElem('methods'); }
	fCodeSource() { return this.getElem('codeSource'); }

}

//------------------------------------------------------

AlgoLabController;

function buildController() {
	//one and only controller (global)
	// eslint-disable-next-line no-global-assign
	gController = new AlgoLabController();
}


