
const log = console.log;

//framework - argument here should be name of this test case
addTestCase("SourceRunnerTest");

let gTemp = null;	//global temp var for testing

class SourceRunnerTest extends TestCase {

	//-----------------------------
	//Constructors

	constructor() {
		super();
		this.model = null;	//or "unit"
		this.holder = null;
	}

	//-----------------------------
	//Helpers

	addToHolder(aElem) {
		this.holder.add(aElem);
	}

	//-----------------------------
	//Test Methods

	test_validationNoErrors() {
		let src, model;
		src = `console.log("Hello")`;
		model = SourceRunner.fromSource(src, AlgoLabModel.getLogSwitchFct());
		model.validate();
		this.assertFalse(model.hasError());
	}

	/*test_withError_BLOB() {
		//This is the only way we get a blob to work -- by putting timeout in > 0
		//WE DISCARDED BLOB APPROACH, SO THIS IS NOT USED.
		let src, errorHandler, f;
		this.holder = [];
		src = `badCallHere(); \n console.log("Hello")`;
		errorHandler = error => this.addToHolder(error);
		this.model = SourceRunner.fromSource(src, errorHandler, this.preFunction(), this.postFunction());
		this.model.load();
		f = (tester) => {
			let errors;
			tester.model.run();
			errors = this.model.getError();
			tester.assertTrue(errors && errors.length > 0);
		};
		setTimeout(f, 0, this);
	}*/

	test_withError_Line1() {
		let src, model, err;
		src =
`badCallHere();
console.log("Hello")`;
		model = SourceRunner.fromSource(src, AlgoLabModel.getLogSwitchFct());
		model.validate();
		this.assertTrue(model.hasError());
		err = model.getError();
		this.assertTrue(err.getMessage().includes('badCallHere'));
		this.assertEquals(1, err.getLineNumber());
	}

	test_withError_Line3() {
		let src, model, err;
		src =
`console.log("Line 1");
console.log("Line 2");
badCallHere();
console.log("Hello")`;
		model = SourceRunner.fromSource(src, AlgoLabModel.getLogSwitchFct());
		model.validate();
		this.assertTrue(model.hasError());
		err = model.getError();
		this.assertTrue(err.getMessage().includes('badCallHere'));
		this.assertEquals(3, err.getLineNumber());
		this.assertEquals('badCallHere();', err.getLineText());
	}

	test_withError_Line3_withComments_1() {
		let src, model, err;
		src =
`console.log("Line 1");
//Foo
console.log("Line 3");
//Boo
badCallHere();
console.log("Hello")`;
		model = SourceRunner.fromSource(src, AlgoLabModel.getLogSwitchFct());
		model.validate();
		this.assertTrue(model.hasError());
		err = model.getError();
		this.assertTrue(err.getMessage().includes('badCallHere'));
		this.assertEquals(5, err.getLineNumber());
	}

	test_withError_Line3_withComments_2() {
		let src, model, err;
		src =
`console.log("Line 1");
//Foo
console.log("Line 3");
/*hey
wow

oh ya*/
console.log("Line 8");
badCallHere();
console.log("Hello")`;
		model = SourceRunner.fromSource(src, AlgoLabModel.getLogSwitchFct());
		model.validate();
		this.assertTrue(model.hasError());
		err = model.getError();
		this.assertTrue(err.getMessage().includes('badCallHere'));
		this.assertEquals(9, err.getLineNumber());
	}


	test_withError_Line3_withComments_ErrorLastLine() {
		let src, model, err;
		src =
`console.log("Line 1");
//Foo
console.log("Line 3");
/*hey
wow

oh ya*/
console.log("Line 8");
badCallHere();`;
		model = SourceRunner.fromSource(src, AlgoLabModel.getLogSwitchFct());
		model.validate();
		this.assertTrue(model.hasError());
		err = model.getError();
		this.assertTrue(err.getMessage().includes('badCallHere'));
		this.assertEquals(src.split('\n').length, err.getLineNumber());
	}


	test_withSyntaxError_1() {
		let src, model, err;
		src =
`console.log("Line 1");
)
console.log("Line 3");`;
		model = SourceRunner.fromSource(src, AlgoLabModel.getLogSwitchFct());
		model.validate();
		this.assertTrue(model.hasError());
		err = model.getError();
		this.assertTrue(err.getMessage().includes(')'));
		this.assertEquals(2, err.getLineNumber());
	}


	test_withSyntaxError_2() {
		let src, model, err;
		src =
`console.log("Line 1");
console.log("Line 3");
((())`;
		model = SourceRunner.fromSource(src, AlgoLabModel.getLogSwitchFct());
		model.validate();
		this.assertTrue(model.hasError());
		err = model.getError();
		this.assertTrue(err.getMessage().includes('(') || err.getMessage().includes(')'));
		this.assertEquals(3, err.getLineNumber());
	}


	test_run_1() {
		let src, model, okay;
		src = `gTemp = 31416`;
		model = SourceRunner.fromSource(src, AlgoLabModel.getLogSwitchFct());
		okay = model.validate();
		this.assertTrue(okay);
		if (okay) {
			model.run();
			this.assertEquals(31416, gTemp);
		}
	}

	test_run_2() {
		let src, model, okay;
		src =
`
//a little test
let a = 10
let b = 5
gTemp = a + b
/*all done
yep*/`;
		model = SourceRunner.fromSource(src, AlgoLabModel.getLogSwitchFct());
		okay = model.validate();
		this.assertTrue(okay);
		if (okay) {
			model.run();
			this.assertEquals(15, gTemp);
		}
	}

	//-----------------------------
	//Overrides - Static

	static testSelectors() {
		return [
			'test_noErrors',
			'test_withErrors_Line1',
			'test_withErrors_Line3',
			'test_withError_Line3_withComments_1',
			'test_withError_Line3_withComments_2',
			'test_withSyntaxError_1',
			'test_withSyntaxError_2',
			'test_run_1',
			'test_run_2'
		];
	}

}

//-----------------------------------------------------

// if (require.main === module)
// 	MyStringTest.runAll();

SourceRunnerTest;
