
/*
source-runner.js
class: SourceRunner
*/

class SourceRunner {

    //---------------------------------------------------------------------------------\
    //Construction

    static newRunner() {
        return new this();
    }

    static fromSource(src, logSwitch) {
        const runner = new this();
        runner.source = src;
        runner.sourceFileSimulator = null;
        runner.validator = SourceValidator.fromSource(src, logSwitch);
        runner.runtimeErrors = [];
        return runner;
    }

    //---------------------------------------------------------------------------------\
    //Accessors

    getError() {
        return this.validator.getError();
    }

    hasError() {
        return this.validator.hasError();
    }

    getRuntimeErrors() { return this.runtimeErrors; }

    addRuntimeError(errorMessage) {
        this.getRuntimeErrors().add(errorMessage);
    }

    //---------------------------------------------------------------------------------\
    //Public

    validate() {
        this.validator.validate();
        return !this.hasError();
    }

    load() {
        if (!this.validator.isDone())
            this.validate();
        if (this.hasError())
            return false;
        this.loadSource();
        return true;
    }

    run() {
        this.runSource();
        //this.runAsFile();
    }

    //---------------------------------------------------------------------------------\
    //Private

    validateSyntax() {
        //private
        //We run as (mock) file, as it allows catching errors (vanilla source code does not)
        /*unfortunately, the blob does not give us load errors for try/catch
        thus a little kludge and we wrap the entire script in function dummyOuterFct
        and run it -- NOTE WELL -- dummyOuterFct call must be in timeoutFct (i.e., queued)
        that ensures code is loaded*/
        //CHANGE -- we hoped to catch errors in a blob -- at least we can get line-numbers
        //of non-syntax errors (blob does NOT help with line #'s of  syntax errors). However
        //blob is in ui effectively (i.e., adding a HTML element) -- and ui thread is flakey,
        //e.g. need to do a setTimeout for some set time (not 0) and the time is just
        //a guess. Thus => we comment out "if (ex instanceof SyntaxError)" below
        try {
            this.logSwitch(false);
            (new Function(this.source))();
        } catch (ex) {
            //if (ex instanceof SyntaxError)
            this.errors.add('SYNTAX ERROR -- ' + ex.toString());
        } finally {
            this.logSwitch(true);
        }
    }

    getSourceWrapped() {
        return `function dummyOuterFct() { \n ${this.source} \n }`;
    }

    loadSource() {
        //js does not seem to generate load errors (load is on-demand so not triggered
        //until run (it seems), but we catch here just in case
        try {
            this.sourceFileSimulator = new SourceFileSimulator(this.getSourceWrapped());
            //this.sourceFileSimulator = new SourceFileSimulator(this.source);
            this.sourceFileSimulator.load();
        } catch (ex) {
            this.addRuntimeError('LOAD ERROR -- ' + ex.toString());
        }
    }

    //-------------------------------------------------------------
    //private

    runSource() {
        //poormans "backup"
        copyToClipboard(this.source);
        try {
            let fct;
            fct = new Function(this.source);
            fct();
        } catch (ex) {
            const error = SourceErrorParser.fromException(ex, this.source);
            const errorString = error.getErrors().join(newline());
            this.addRuntimeError(errorString);
        }
    }

    runAsFile() {
        //this has problematic timing (as it involves loading into messy UI, we
        //prefer "runSource")
        try {
            dummyOuterFct();
        } catch (ex) {
            const error = SourceErrorParser.fromException(ex, this.source);
            const errorString = error.getErrors().join(newline());
            this.addRuntimeError(errorString);
        } finally {
            this.sourceFileSimulator.unload();
        }
    }

}

//-------------------------------------------------------------
SourceRunner;